from .base_controller import BaseController
from .message_retrieval import MessageRetrieval
from .media_controller import MediaController
import pandas as pd
from models.message import Message
from models.timestamp_tracker import TimestampTracker
from models.base import Session
from utils.api_service import APIService
import logging
import json
from datetime import datetime

logger = logging.getLogger(__name__)

class MessageProcessor:
    @staticmethod
    def process_new_messages():
        logger.info("=== Starting new message processing cycle ===")
        session = Session()

        try:
            last_timestamp = TimestampTracker.get_last_timestamp()
            current_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

            logger.info(f"Current timestamp in DB: {last_timestamp}")

            logger.info(f"Fetching messages from {last_timestamp or 'beginning'} to {current_time}")
            new_messages = APIService.fetch_messages(last_timestamp, current_time)

            if not new_messages:
                logger.info("No new messages found in this cycle")
                return 0

            logger.info(f"Retrieved {len(new_messages)} new messages from API")

            # Process messages into DataFrame
            processed_data = MessageProcessor._process_message_data(new_messages, session)

            if not processed_data.empty:
                # Save messages one by one and get counts
                saved_count, duplicate_count = Message.save_messages(processed_data)

                # Update timestamp tracker with latest message timestamp regardless of saved count
                latest_timestamp = processed_data["timestamp"].max()
                def update_timestamp():
                    timestamp_tracker = session.query(TimestampTracker).with_for_update().first()
                    if not timestamp_tracker:
                        timestamp_tracker = TimestampTracker(last_timestamp=latest_timestamp)
                        session.add(timestamp_tracker)
                    else:
                        timestamp_tracker.last_timestamp = latest_timestamp
                    session.commit()

                BaseController._execute_with_retry(session, update_timestamp)
                logger.info(f"Timestamp tracker updated to {latest_timestamp}")

                if latest_timestamp:
                    logger.info(f"Updating timestamp to: {latest_timestamp}")
                    TimestampTracker.update_last_timestamp(latest_timestamp)

                return saved_count

            return 0

        except Exception as e:
            logger.error(f"Error processing new messages: {str(e)}")
            raise
        finally:
            session.close()

    @staticmethod
    def _process_message_data(messages_data, session):
        if not messages_data:
            logger.error("Received empty messages_data")
            return pd.DataFrame([])

        try:
            processed_data_list = []

            for msg in messages_data:
                try:
                    if msg is None:
                        logger.warning("Skipping None message")
                        continue

                    # Process single message
                    processed_message = MessageProcessor._process_single_message(msg, session)
                    if processed_message:
                        # Save the message and handle media download synchronously
                        messages_df = pd.DataFrame([processed_message])
                        saved_count, duplicate_count = Message.save_messages(messages_df)

                        # Only proceed to next message if this one was successfully saved
                        if saved_count > 0:
                            processed_data_list.append(processed_message)
                            logger.info(f"Successfully saved message ID: {processed_message.get('message_id')}")
                        else:
                            if duplicate_count > 0:
                                logger.warning(f"Message not saved (duplicate detected), skipping to next message")
                            else:
                                logger.warning(f"Message not saved (unknown reason), skipping to next message")
                                # Log more details about the message for debugging
                                logger.debug(f"Message details: type={processed_message.get('message_type')}, "
                                           f"chat_id={processed_message.get('chat_id')}, "
                                           f"body_length={len(processed_message.get('body', '')) if processed_message.get('body') else 0}")

                except Exception as e:
                    logger.error(f"Error processing individual message: {str(e)}", exc_info=True)
                    continue

            return pd.DataFrame(processed_data_list)

        except Exception as e:
            logger.error(f"Error in _process_message_data: {str(e)}", exc_info=True)
            return pd.DataFrame([])

    @staticmethod
    def _process_single_message(msg, session):
        """Process a single message and return the processed data"""
        try:
            logger.info(f"Raw message data: {json.dumps(msg, default=str)[:200]}")  # Log first 200 chars
            if not msg or not isinstance(msg, dict):
                logger.warning("Invalid message format")
                return None

            # Check for unknown message type
            message_type = msg.get('messageType', '').lower()
            valid_types = {'chat', 'video', 'image'}
            if message_type not in valid_types:
                logger.warning(f"Invalid message type '{message_type}', skipping message: {msg.get('messageId')}")
                return None

            # Get or create author
            author_name = msg.get('author')
            author_number = msg.get('number')
            author_id = MessageRetrieval._get_or_create_author(author_name, author_number)

            # Process timestamp
            timestamp = msg.get('timestamp')
            try:
                timestamp = pd.to_datetime(timestamp) if timestamp else datetime.now()
            except Exception as e:
                logger.error(f"Timestamp parse error: {e}")
                timestamp = datetime.now()

            # Handle media
            local_url = None
            file_size = None
            file_hash = None

            logger.info(f"Processing message type: {message_type}")

            if message_type in ['image', 'video']:
                # Extract media info
                media_info = msg.get('mediaInfo', {})
                file_hash = media_info.get('fileHash')
                file_size = media_info.get('fileSize')

                # Check for duplicate first
                duplicate = Message.find_duplicate_message(
                    message_type=message_type,
                    file_hash=file_hash
                )

                if duplicate:
                    logger.info(f"Found duplicate media with hash {file_hash}, reusing file at {duplicate.local_url}")
                    local_url = duplicate.local_url
                else:
                    logger.info(f"Attempting media download for message type: {message_type}")
                    local_url, file_size, file_hash = MediaController._handle_media_download(
                        msg, message_type, timestamp
                    )
                    if not local_url:
                        local_url = msg.get('localUrl')  # Fallback to original URL

            return {
                'message_id': msg.get('messageId'),
                'body': (msg.get('body') or '').strip(),
                'caption': (msg.get('caption') or '').strip(),
                'message_type': message_type,
                'timestamp': timestamp,
                'chat_id': msg.get('chat', {}).get('id') or msg.get('chatId'),
                'author_id': author_id,
                'number': author_number,
                'local_url': local_url,
                'file_size': file_size,
                'file_hash': file_hash
            }

        except Exception as e:
            logger.error(f"Message processing error: {str(e)}", exc_info=True)
            return None


